import os
from os.path import join, dirname, splitext, isabs
import re
from typing import Callable, Awaitable, Dict, Any, List, Set, Union, Tuple
from logging import getLogger, Logger
import json
import logging
import glob
import uuid
from queue import Queue
from os.path import exists
from pathlib import Path

from aiohttp import web, WSMsgType
from openai import OpenAI

from agi_green.dispatcher import Protocol, format_call, protocol_handler
from agi_green.config_namespace import DictNamespace

here = dirname(__file__)
logger = logging.getLogger(__name__)
log_level = os.getenv('LOG_LEVEL', 'WARNING').upper()
logging.basicConfig(level=log_level)


text_content_types = {
    '.html': 'text/html',
    '.js': 'application/javascript',
    '.map': 'application/json',
    '.css': 'text/css',
    '.txt': 'text/plain',
    '.md': 'text/markdown',
}

class HTTPServerProtocol(Protocol):
    '''
    http server (or https if ssl_context is provided)

    If ssl_context is provided, use https, and launch a http->https redirect server
    Otherwise, use http

    Either way, our naming convention assumes http (i.e. the class name and protocol_id)
    '''

    protocol_id: str = 'http'

    def __init__(self, parent:Protocol, host:str='0.0.0.0', port:int=8000, ssl_context=None, redirect=None):
        super().__init__(parent)

        self.host = host
        self.port = port
        self.redirect = redirect
        self.ssl_context = ssl_context
        self.app:web.Application = None
        self.runner:web.AppRunner = None
        self.site:web.TCPSite = None
        self.session_class = self.dispatcher.session_class
        self.sessions:Dict[str, Protocol] = {}

    async def http_to_https_redirect(self, request):
        assert self.ssl_context is not None, "SSL context must be set for HTTPS redirect"
        https_location = f'https://{request.host}{request.rel_url}'
        raise web.HTTPMovedPermanently(https_location)

    def get_or_create_session(self, request):
        session_id = request.cookies.get('SESSION_ID')
        new_session_id = None

        if not session_id:
            new_session_id = session_id = str(uuid.uuid4())

        try:
            session = self.sessions[session_id]
        except KeyError:
            session = None

        if session is None:
            session:Protocol = self.session_class(self, session_id=session_id)
            self.sessions[session_id] = session
            self.add_task(session.run())

        # Set the subdomain in the session context
        host = request.host.split(':')[0]  # Remove port if present
        session.context.subdomain = host

        return session, new_session_id

    async def handle_http_request(self, request:web.Request):
        logger.info(f"HTTP Request received: {request.method} {request.path}")
        session, new_session_id = self.get_or_create_session(request)
        http:HTTPSessionProtocol = session.get_protocol('http')

        # Convert headers to a simple dict for message passing
        headers = {k: v for k, v in request.headers.items()}

        response:web.StreamResponse|None = await http.handle_request(request, headers=headers)
        self.context.host = request.host

        if new_session_id:
            logger.info(f'New session: {new_session_id}')
            if response is None:
                logger.error(f'Request failed on new session {new_session_id} on http request:')
                logger.error(f'  {request}')
            else:
                response.set_cookie('SESSION_ID', new_session_id, max_age=60*60*24*365)
                logger.info(f'New session: {new_session_id}')
        else:
            logger.info(f'Existing session: {session.context.session_id}')

        return response

    async def handle_websocket_request(self, request:web.Request):
        socket = web.WebSocketResponse()
        await socket.prepare(request)
        session, new_session_id = self.get_or_create_session(request)
        socket.id = request.query['socket_id']

        # Convert headers to a simple dict for message passing
        headers = {k: v for k, v in request.headers.items()}
        logger.debug(f"WebSocket connection headers: {headers}")

        # Pass the socket to the ws protocol's connect handler
        ws = session.get_protocol('ws')
        await ws.handle_mesg('connect', socket=socket, headers=headers)

        async for msg in socket:
            logger.info(f'ws {msg.type}, {msg.data}')
            if msg.type == WSMsgType.TEXT:
                data = json.loads(msg.data)
                # Handle the message
                await ws.handle_mesg(**data)
            elif msg.type == WSMsgType.ERROR:
                logger.error('ws connection closed with exception %s' % socket.exception())
            else:
                logger.info('ws {msg.type}')

        # Handle disconnect
        await ws.handle_mesg('disconnect', socket=socket)

        if new_session_id:
            logger.error(f'Unexpected new session on ws message: {self} {new_session_id}')

        return socket


    async def run(self):
        self.add_task(super().run())

        self.app = web.Application(client_max_size=10_000_000_000)  # 10GB limit to match websocket
        logger.info(f'web.Application(client_max_size=10_000_000_000)')
        # on_http_* methods are handled by HTTPSessionProtocol
        #handle_websocket_request
        self.app.router.add_get('/ws', self.handle_websocket_request)  # Delegate WebSocket connections
        self.app.router.add_get('/{filename:.*}', self.handle_http_request)
        self.app.router.add_post('/{filename:.*}', self.handle_http_request)
        self.app.router.add_get('/', self.handle_http_request, name='index')

        # Check if SSL context is provided for HTTPS
        if self.ssl_context:
            # HTTPS server setup
            self.runner = web.AppRunner(self.app)
            await self.runner.setup()
            self.site = web.TCPSite(self.runner, self.host, self.port, ssl_context=self.ssl_context)
            logger.info(f'Serving https://{self.host}:{self.port}')
            await self.site.start()

            if self.redirect:
                # Additional HTTP server for redirecting to HTTPS
                redirect_app = web.Application()
                redirect_app.router.add_get('/{tail:.*}', self.http_to_https_redirect)
                redirect_runner = web.AppRunner(redirect_app)
                await redirect_runner.setup()
                redirect_site = web.TCPSite(redirect_runner, self.host, self.redirect)
                logger.info(f'Starting HTTP redirect server on http://{self.host}:{self.redirect}')
                await redirect_site.start()
        else:
            # HTTP server setup
            self.runner = web.AppRunner(self.app)
            await self.runner.setup()
            self.site = web.TCPSite(self.runner, self.host, self.port)
            await self.site.start()
            logger.info(f'serving http://{self.host}:{self.port}')
            logger.info(f'{self.app.router}')


    async def close(self):
        # Stop the aiohttp site
        if self.site:
            await self.site.stop()

        # Shutdown and cleanup the aiohttp app
        if self.app:
            await self.app.shutdown()
            await self.app.cleanup()

        # Finally, cleanup the AppRunner
        if self.runner:
            await self.runner.cleanup()

        await super().close()


class HTTPSessionProtocol(Protocol):
    '''
    Session http protocol handler
    This is instantiated for each user who connects to the server
    '''

    protocol_id: str = 'http'

    def __init__(self, parent:Protocol):
        super().__init__(parent)
        self.static = [join(here, 'static'), join(here, 'frontend', 'dist')]
        self.static_handlers:List[Callable] = []

        for static_dir in self.static:
            if not exists(static_dir):
                logger.warning(f'Static directory {static_dir}: does not exist')
                logger.warning('Did you forget to run "npm run build" in the frontend directory?')

    def add_static(self, path:str, index:int=None):
        'add static directory'
        logger.info(f'Adding static directory {path}')

        if not exists(path):
            logger.warning(f'Static directory {path}: does not exist')

        if index is None:
            self.static.append(path)
        else:
            self.static.insert(index, path)

    def add_static_handler(self, handler:Callable, index:int=None):
        'add static handler'
        if index is None:
            self.static_handlers.append(handler)
        else:
            self.static_handlers.insert(index, handler)

    def find_static(self, filename:str):
        if '*' in filename:
            # Handle glob pattern
            for static_dir in self.static:
                file_path = os.path.join(static_dir, filename)
                matches = glob.glob(file_path)
                if matches:
                    if len(matches) > 1:
                        # Return the largest file which is likely the main bundle
                        return max(matches, key=os.path.getsize)
                    return matches[0]
        else:
            # Normal exact match
            for static_dir in self.static:
                file_path = os.path.join(static_dir, filename)
                if os.path.isfile(file_path):
                    return file_path

        return None

    def find_static_glob(self, filename:str):
        files = []
        for static_dir in self.static:
            file_path = os.path.join(static_dir, filename)
            files.extend(glob.glob(file_path))
        return files

    def index_md(self):
        index_file = join(here, 'static', 'docs', 'index.md')
        files = self.find_static_glob('docs/*.md')
        newest_file = max(files, key=os.path.getmtime)

        if newest_file != index_file:
            if index_file in files:
                files.remove(index_file)
            files.sort()

            with open(index_file, 'w') as f:
                f.write(f'<!-- This index is generated by {__file__} - edits will be lost. -->\n\n')
                f.write('| File | Description |\n')
                f.write('| ---- | ----------- |\n')
                for file in files:
                    with open(file, 'r') as f2:
                        s = f2.read()
                        # find first markdown header
                        m = re.search(r'^#+\s+(.*)', s, re.MULTILINE)
                        if m:
                            header = m.group(1)
                            base = os.path.basename(file).replace('.md','')
                            f.write(f'| [**{base}**](/docs/{base}) | *{header}* |\n')


        if not exists(index_file):
            return None
        return index_file

    async def handle_request(self, request:web.Request, headers:dict=None):
        data = DictNamespace()
        url = str(request.url)

        if request.method == 'POST':
            data.update(await request.post())

        data.update(request.query)

        # Send request to all protocols before serving static files
        data.request_url = url
        data.request_method = str(request.method)
        data.headers = headers

        # First try http_request handlers
        response = await self.handle_mesg('request', **data)

        if response is not None:
            if isinstance(response, dict):
                if response.get('__break__', False):
                    return response.get('response')
                return response.get('response', response)
            return response

        if request.url.name and data and '.' not in request.url.name:
            # it might be a command, so try sending it to http protocol handlers
            data.request_url = url
            data.request_method = str(request.method)
            data.headers = headers  # Add headers to data
            cmd = request.url.path[1:].replace('/', '_')

            response = await self.handle_mesg(cmd, **data)

            if response is not None:
                if isinstance(response, web.Response):
                    return response
                elif isinstance(response, str):
                    return web.Response(text=response, content_type='text/plain')
                else:
                    logger.error(f'{data.request_method} {cmd} invalid return type')
                    return web.Response(text='invalid return type', content_type='text/plain')
            else:
                if data.request_method == 'POST':
                    logger.error(f'POST {cmd} no response (did the handler forget to return something?)')
                    return web.Response(text=f'no response to POST {cmd}', content_type='text/plain')

        if request.method == 'GET':
            filename = request.match_info['filename'] or 'index.html'

            query = request.query.copy()

            if filename == 'docs':
                file_path_md = self.index_md()
                filename = 'docs/index'
            else:
                # check for filename+'.md' and serve that instead with query: view=render
                file_path_md = self.find_static(filename+'.md')

            if file_path_md is not None:
                query.add('view','render')
                file_path = file_path_md
                filename = filename+'.md'
            else:
                file_path = self.find_static(filename)

            if file_path is None:
                for h in self.static_handlers:
                    file_path = h(filename)
                    if file_path is not None:
                        break
                else:
                    return web.HTTPNotFound()

            ext = os.path.splitext(filename)[1]
            content_type = text_content_types.get(ext, None) # None means binary

            if content_type == 'text/markdown':
                format = query.get('view', 'raw')

                if format == 'raw':
                    return web.FileResponse(file_path)

                if not os.path.exists(file_path) or not os.path.isfile(file_path):
                    raise web.HTTPNotFound()

                with open(file_path, 'r') as f:
                    content = f.read()

                # serve the index.html file. The open_md message will populate the md viewer
                file_path = self.find_static('index.html')

                # queue up the message (will be queued until after the websocket is connected)
                await self.send('ws', 'open_md', name=filename, content=content, viewmode='render')

                return await self.serve_file(file_path)

            else:
                response = await self.serve_file(file_path)

        # After getting any response, pass through http_response handler
        if response is not None:
            await self.handle_mesg('response', request=request, response=response, **data)

        return response


    @staticmethod
    async def serve_file(file_path):
        logger.info(f"Serving file: {file_path}")
        response = web.FileResponse(file_path)

        # Add cache control headers
        response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'
        response.headers['Pragma'] = 'no-cache'
        response.headers['Expires'] = '0'

        # Manually set Content-Type for .js.map files
        if file_path.endswith('.js.map'):
            response.content_type = 'application/json'
        elif file_path.endswith('.js'):
            response.content_type = 'application/javascript'

        print(f'{file_path} => {response.content_type}')

        return response



import os
from os.path import join, dirname, splitext, isabs
import time
from typing import Callable, Awaitable, Dict, Any, List, Set, Union, Tuple
from logging import getLogger, Logger
import json
import asyncio
import logging
from os.path import exists
import uuid

from aiohttp import web

from agi_green.dispatcher import Protocol, format_call, protocol_handler

here = dirname(__file__)
logger = logging.getLogger(__name__)
log_level = os.getenv('LOG_LEVEL', 'WARNING').upper()
logging.basicConfig(level=log_level)


WS_PING_INTERVAL = 20

class WebSocketProtocol(Protocol):
    '''
    Websocket session
    '''
    protocol_id: str = 'ws'

    def __init__(self, parent:Protocol):
        super().__init__(parent)
        self.sockets: Set[web.WebSocketResponse] = set()
        self.socket_states: Dict[str, Dict] = {}
        self.pre_connect_queue = []

    async def ping_loop(self, socket: web.WebSocketResponse):
        'ping the websocket to keep it alive'
        last_pong_time = time.time()

        while socket in self.sockets:
            try:
                await socket.ping()
            except ConnectionResetError as e:
                logger.error(f'ws connection reset (closing) {e} {self.dispatcher.session_id}')
                self.sockets.discard(socket)
                break
            await asyncio.sleep(WS_PING_INTERVAL)

    async def do_send(self, cmd: str, socket_id: str = None, **kwargs):
        'send ws message to specific socket or all connected browsers'
        kwargs['cmd'] = cmd

        if not self.sockets:
            logger.info(f'queuing ws: {format_call(cmd, kwargs)}')
            self.pre_connect_queue.append(kwargs)
            return

        try:
            s = json.dumps(kwargs)
            logger.info(f'Attempting to send WebSocket message: {s}')
        except Exception as e:
            logger.error(f'ws send error: {e})')
            logger.error(f'ws send error: {kwargs}')
            return

        dead_sockets = set()
        target_sockets = [s for s in self.sockets if socket_id is None or getattr(s, 'id', None) == socket_id]

        for socket in target_sockets:
            try:
                logger.info(f'Sending to socket {getattr(socket, "id", "unknown")}: {s}')
                await socket.send_str(s)
                logger.info(f'Successfully sent to socket {getattr(socket, "id", "unknown")}')
            except Exception as e:
                logger.error(f'ws send error: {e} (removing socket)')
                dead_sockets.add(socket)

        self.sockets -= dead_sockets

    @protocol_handler(priority=0)
    async def on_ws_connect(self, socket: web.WebSocketResponse):
        """Handle WebSocket connection"""

        if not socket.id:
            logger.error('No socket.id (should be set by HTTPServerProtocol.handle_websocket_request)')
            raise ValueError('No socket.id')

        self.sockets.add(socket)
        self.add_task(self.ping_loop(socket))

        return {'socket': socket}

    @protocol_handler
    async def on_ws_disconnect(self, socket: web.WebSocketResponse):
        self.sockets.discard(socket)
        # Keep state for potential reconnect
        # Add cleanup after timeout
        self.add_task(self._cleanup_socket_state(socket.id))

    async def _cleanup_socket_state(self, socket_id: str):
        """Remove socket state if no reconnect within timeout"""
        await asyncio.sleep(30)  # Adjust timeout as needed
        if socket_id in self.socket_states:
            state = self.socket_states[socket_id]
            if time.time() - state['last_ping'] > 30:
                del self.socket_states[socket_id]

    async def handle_message(self, socket: web.WebSocketResponse, data: Dict):
        """Handle incoming WebSocket message"""
        # Debug logging
        logger.info(f"Received message with data: {data}")
        await super().handle_message(socket, data)




import os
from pathlib import Path
import asyncio
import logging
from typing import Dict, Union, Set
from watchfiles import awatch
from agi_green.dispatcher import Protocol

logger = logging.getLogger(__name__)

class FileProtocol(Protocol):
    """Protocol for monitoring file changes"""
    protocol_id: str = 'file'

    def __init__(self, parent: Protocol):
        super().__init__(parent)
        self.monitored_paths: Dict[str, Set[str]] = {}
        self.watch_tasks: Dict[str, asyncio.Task] = {}

    def _normalize_path(self, path: Union[str, Path]) -> Path:
        """Convert string or Path to normalized Path"""
        return Path(path).resolve()

    async def _watch_directory(self, dir_path: str):
        """Watch a directory for changes"""
        try:
            async for changes in awatch(dir_path):
                for _, path_str in changes:
                    path_str = str(Path(path_str).resolve())
                    # Check all channels for matching paths
                    for channel, paths in self.monitored_paths.items():
                        if path_str in paths:
                            await self.handle_mesg(channel, path=path_str)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            logger.error(f"Error watching directory {dir_path}: {e}")

    def monitor(self, channel: str, path: Union[str, Path]):
        """Set up file monitoring for the given channel and path"""
        path = self._normalize_path(path)
        path_str = str(path)
        dir_path = str(path.parent)

        # Initialize channel set if needed
        if channel not in self.monitored_paths:
            self.monitored_paths[channel] = set()

        # Add path to monitored set
        self.monitored_paths[channel].add(path_str)

        # Set up directory watching if not already watched
        if dir_path not in self.watch_tasks:
            try:
                task = self.add_task(self._watch_directory(dir_path))
                self.watch_tasks[dir_path] = task
                logger.info(f"Monitoring directory: {dir_path}")
            except Exception as e:
                logger.error(f"Failed to monitor directory {dir_path}: {e}")

        logger.info(f"Monitoring file {path_str} on channel {channel}")

    def unmonitor(self, channel: str, path: Union[str, Path] = None):
        """Stop monitoring file(s) for the given channel"""
        if channel not in self.monitored_paths:
            return

        if path is None:
            # Remove all monitored paths for this channel
            self.monitored_paths[channel].clear()
        else:
            # Remove specific path
            path_str = str(self._normalize_path(path))
            self.monitored_paths[channel].discard(path_str)

        # If channel has no more monitored paths, remove it
        if not self.monitored_paths[channel]:
            del self.monitored_paths[channel]

        # Clean up any directories that are no longer needed
        needed_dirs = {str(Path(p).parent) for paths in self.monitored_paths.values() for p in paths}
        for dir_path in list(self.watch_tasks.keys()):
            if dir_path not in needed_dirs:
                self.watch_tasks[dir_path].cancel()
                del self.watch_tasks[dir_path]

    async def close(self):
        """Clean up resources"""
        for task in self.watch_tasks.values():
            task.cancel()
        await super().close()



from datetime import datetime, timezone
import re
import bcrypt
import secrets
from agi_green.agi_db import get_collection
from agi_green.dispatcher import Protocol, protocol_handler

re_session_id = re.compile(r'^[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}$')

login_form = '''```form-yaml
id: login
schema:
  h2:
    type: static
    tag: h3
    content: ARIA Login
  user:
    type: text
    label: Screen Name or Email
  password:
    type: text
    inputType: password
    label: Password
  submit:
    type: button
    buttonLabel: Login
    submits: true
```'''

reg_form = '''```form-yaml
id: regform
schema:
  h2:
    type: static
    tag: h3
    content: ARIA Registration
  screen_name:
    type: text
    label: Screen Name
  email:
    type: text
    label: Email
  password:
    type: text
    inputType: password
    label: Password
  confirm:
    type: text
    inputType: password
    label: Confirm Password
  submit:
    type: button
    buttonLabel: Register
    submits: true
```'''

class UserProtocol(Protocol):
    protocol_id = "user"

    def __init__(self, parent:Protocol):
        self.is_authenticated = False
        super().__init__(parent)

    async def create_confirmed_user(self, email: str, screen_name: str, password: str) -> str:
        'create a new user and authenticate them'
        result = await self.create_user(email, screen_name, password)
        if result == 'created':
            # Set email as confirmed
            users_collection = get_collection("users")
            await users_collection.update_one(
                {"email": email}, 
                {"$set": {"email_confirmed": True}}
            )
        return result   

    async def create_user(self, email: str, screen_name: str, password: str) -> str:
        '''create a new user, return status:
        "created": new unconfirmed user created
        "updated": existing unconfirmed user updated
        "exists": not created because of an existing confirmed user with same email or screen_name
        '''
        users_collection = get_collection("users")
        hashed_password = bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt())
        confirmation_code = secrets.token_urlsafe()
        new_user = {
            "email": email,
            "screen_name": screen_name,
            "password": hashed_password,
            "email_confirmed": False,
            "confirmation_code": confirmation_code,
            "session_id": self.context.session_id,
            "created_at": datetime.utcnow().isoformat(),  # Current timestamp in ISO 8601 format
            "last_login": None,  # You can update this field when the user logs in
            "roles": ["user"],  # This could be useful for role-based access control (RBAC)
            "profile": {
                "avatar_url": None,
                "bio": None,
                "location": None,
            },
            "status": "active"  # Could be "active", "inactive", "suspended", etc.
        }

        # Check for existing users with the same email or screen_name
        existing_user = await users_collection.find_one({
            "$or": [
                {"email": email},
                {"screen_name": screen_name}
            ]
        })

        if existing_user:
            if existing_user['email_confirmed']:
                # If there's an existing confirmed user with the same email or screen_name, raise an error
                return 'exists'
            else:
                # If there's an existing unconfirmed user, replace it with the new unconfirmed user details
                await users_collection.replace_one({
                    "_id": existing_user['_id']
                }, new_user)
                result = 'updated'
        else:
            # If no existing user matches, create a new unconfirmed user document
            await users_collection.insert_one(new_user)
            result = 'created'

        # Common code for sending confirmation email, executed for both new and replaced unconfirmed users
        conf_email = self.config.conf_email.copy()
        conf_email.update(self.config.email)
        sender = conf_email.pop("sender")  # sender must be first arg (as cmd)

        # Send confirmation email
        await self.send('email', sender, recipient=email, to_name=screen_name,
                        code=confirmation_code, email=email, **conf_email)

        self._confirmation_passwd = password # save password for later
        self._confirmation_user = screen_name

        return result

    async def validate_password(self, password: str) -> bool:
        'validate password against password policy'
        if len(password) < 8:
            await self.send('ws', 'append_chat', author='info', content="Password must be at least 8 characters long.")
            return False
        return True

        # TODO: Use an agent to validate password

    @protocol_handler
    async def on_mq_confirm_user(self):
        'user has confirmed email, automatically log in user'

        if hasattr(self, '_confirmation_passwd'):
            await self.send('ws', 'append_chat', author='info', content=f"Email confirmed successfully. Logging in...")
            success = await self.authenticate(self._confirmation_user, self._confirmation_passwd)
            del self._confirmation_passwd

            if success:
                if not self.is_authenticated:
                    self.is_authenticated = True
                    await self.send('user', 'auth', success=True)

            else:
                await self.send('ws', 'append_chat', author='info', content=f"Automatic login failed. Please log in.\n{login_form}")
        else:
            await self.send('ws', 'append_chat', author='info', content=f"Email confirmed successfully. Please log in.\n{login_form}")


    async def authenticate(self, email_or_name: str, password: str) -> bool:
        '''
        authenticate user by email_or_name and password
        return True if authenticated, False otherwise

        Note that this method immediately returns True if the user is already logged in.
        '''
        # if already logged in, return True
        if (email_or_name == self.context.user.screen_name or
            email_or_name == self.context.user.email or
            email_or_name == self.context.user.session_id):

            if not self.is_authenticated:
                self.is_authenticated = success
                await self.send('user', 'auth', success=True)

            return True

        users_collection = get_collection("users")
        sessions_collection = get_collection("sessions")

        if re_session_id.match(email_or_name):
            user = await users_collection.find_one({"session_id": email_or_name})
            if user:
                if not user.get("email_confirmed", False):
                    return False

                # migrate session_id to user_id, removing obsolete session_id
                sessions_collection.update_one({"session_id": email_or_name}, {
                    "$set": {
                        "last_login": datetime.now(timezone.utc).isoformat(),
                        "user_id": user["_id"]
                    },
                    "$inc": {"logins": 1}
                }, upsert=True)
                await self.send('ws', 'append_chat', author='info', content=f'Welcome back {user["screen_name"]}')
                success = True
            else:
                session = await sessions_collection.find_one({"session_id": email_or_name})
                if session:
                    user = await users_collection.find_one({"_id": session["user_id"]})
                    success = True
                    await self.send('ws', 'append_chat', author='info', content=f'Welcome back {user["screen_name"]}')
                else:
                    return False # session_id not found, so return False silently
        else:
            # user can be either email or screen_name
            if "@" in email_or_name:
                user = await users_collection.find_one({"email": email_or_name})
            else:
                user = await users_collection.find_one({"screen_name": email_or_name})

                if user and not user['email_confirmed'] == True:
                    await self.send('ws', 'append_chat', author='info', content="Please confirm your email before logging in.")
                
                return False

            if user and bcrypt.checkpw(password.encode('utf-8'), user["password"]):
                success = True
                await self.send('ws', 'append_chat', author='info', content=f'Login successful. Welcome {user["screen_name"]}')

                # add the session to the session collection
                sessions_collection.update_one({"session_id": self.context.session_id}, {
                    "$set": {
                        "last_login": datetime.now(timezone.utc).isoformat(),
                        "user_id": user["_id"]
                    },
                    "$inc": {"logins": 1}
                }, upsert=True)

            else:
                success = False
                await self.send('ws', 'append_chat', author='info', content="Login failed. Please check your credentials and try again.")

        if success:
            if not self.is_authenticated:
                self.is_authenticated = success
                await self.send('user', 'auth', success=success)

            uc = self.context.user
            uc.is_authenticated = True
            uc._deep_update(user)
            await self.send('ws', 'set_user_data', uid=uc.screen_name, name=uc.screen_name)

        else:
            uc = self.context.user
            uc.is_authenticated = False

        return success

    @protocol_handler
    async def on_ws_connect(self):
        'authenticate using session_id'

        #TODO: user opt-in for this feature (not on shared computers)

        # check if session_id is already logged in
        success = await self.authenticate(self.context.session_id, "")
        if success:
            if not self.is_authenticated:
                self.is_authenticated = True
                await self.send('user', 'auth', success=True)
        else:
            await self.send('ws', 'append_chat', author='info', content=f"Please log in.\n{login_form}")

    @protocol_handler
    async def on_cmd_regform(self) -> str:
        '[cmd:regform()] -> form content'
        return reg_form

    @protocol_handler
    async def on_cmd_login(self) -> str:
        '[cmd:login()] -> form content'

        return login_form

    async def confirm_email(self, email: str, confirmation_code: str) -> bool:
        'user clicked on email confirmation link: check code and confirm email'
        users_collection = get_collection("users")
        user = await users_collection.find_one({"email": email, "confirmation_code": confirmation_code})
        if user:
            await users_collection.update_one({"email": email}, {"$set": {"email_confirmed": True}})
            await self.send('mq', 'confirm_user', channel=f'session.'+user['session_id'])
            return True
        return False

    @protocol_handler
    async def on_form_regform(self, email:str, screen_name:str, password:str, confirm:str) -> bool:
        'ws:register?email={email}&screen_name={screen_name}&password={password}'

        if password != confirm:
            await self.send('ws', 'append_chat', author='info', content="Passwords do not match.")
            return False

        if await self.validate_password(password):
            result = await self.create_user(email, screen_name, password)
            if result == 'created':
                await self.get_protocol('mq').subscribe('user.'+screen_name)
                mesg = f"Thank you. To complete your registration, please check your email ({email}) for a confirmation link. If you can't find it, check your spam folder."
            elif result == 'updated':
                await self.get_protocol('mq').subscribe('user.'+screen_name)
                mesg = f"Registration updated. Please check your email ({email}) for a confirmation link. If you can't find it, check your spam folder."
            else:
                mesg = f"A user with the same email or screen name already exists."

            await self.send('mq', 'chat', channel=f'user.{screen_name}', author='info', content=mesg)

            return result
        else:
            return False

    @protocol_handler
    async def on_http_confirm(self, email:str, code:str) -> str:
        'agi.green/confirm?email={email}&code={code}'
        http = self.get_protocol("http")
        success = await self.confirm_email(email, code)
        if success:
            return 'Registration email confirmed.'
        else:
            return 'Registration email failed.'

    @protocol_handler
    async def on_form_login(self, user:str, password:str) -> bool:
        success = await self.authenticate(user, password)
        if success:
            await self.send('ws', 'set_user_data', uid=self.context.user.screen_name, name=self.context.user.screen_name)
            await self.send('ws', 'append_chat', author='info',
                            message=f'Login successful. Welcome {self.context.user.screen_name}')

            if not self.is_authenticated:
                self.is_authenticated = True
                await self.send('user', 'auth', success=True)

        return success



import os
from os.path import join, dirname, splitext, isabs
from typing import Callable, Awaitable, Dict, Any, List, Set, Union, Tuple
from logging import getLogger, Logger
import asyncio
import logging

from aiohttp import web, WSMsgType
from openai import OpenAI

from agi_green.dispatcher import Protocol, format_call, protocol_handler

here = dirname(__file__)
logger = logging.getLogger(__name__)
log_level = os.getenv('LOG_LEVEL', 'WARNING').upper()
logging.basicConfig(level=log_level)

class GPTChatProtocol(Protocol):
    '''
    OpenAI GPT Chat protocol

    This is just a POC: simple async wrapper around the OpenAI API in chat mode.
    Next step is to implement HuggingFace transformers and langchain for more control.
    '''
    protocol_id: str = 'gpt'

    _openai_client: OpenAI = None

    def __init__(self, parent:Protocol):
        super().__init__(parent)
        self.name = 'agi.green'
        self.uid = 'bot'

    @property
    def openai_client(self):
        if GPTChatProtocol._openai_client is None:
            api_key = os.environ.get("OPENAI_API_KEY", None)

            if api_key is None:
                raise Exception("gpt protocol needs OPENAI_API_KEY environment variable to be set")

            GPTChatProtocol._openai_client = OpenAI(api_key=api_key)

        return GPTChatProtocol._openai_client

    async def run(self):
        self.add_task(super().run())

        self.messages = [
            {"role": "system", "content": "You are a helpful assistant."},
        ]

    @protocol_handler
    async def on_ws_form_data(self, cmd:str, data:dict):
        key = data.get('key')
        self.config.set('openai.key', key)
        self.messages.append({"role": "system", "content": "OpenAI API key was just now set by the user."})
        await self.get_completion()

    @protocol_handler
    async def on_ws_connect(self):
        await self.send('ws', 'set_user_data', uid='bot', name='GPT-4', icon='/avatars/agibot.png')
        await self.send('ws', 'set_user_data', uid='info', name='InfoBot', icon='/avatars/infobot.png')

    @protocol_handler
    async def on_mq_chat(self, channel_id:str, author:str, content:str):
        'receive chat message from RabbitMQ'
        if author != self.uid:
            self.messages.append({"role": "user", "content": content})
            task = asyncio.create_task(self.get_completion())

    async def get_completion(self):
        loop = asyncio.get_event_loop()
        content = await loop.run_in_executor(None, self.sync_completion)
        if content:
            await self.send('mq', 'chat', channel='chat.public', author=self.uid, content=content)

    def sync_completion(self):
        try:
            # if env DISABLE_GPT4 is set, skip GPT4 completion
            if os.environ.get('DISABLE_GPT', None):
                logger.info('skipping GPT completion')
            else:
                response = self.openai_client.chat.completions.create(model="gpt-4",
                messages=self.messages)
                return response.choices[0].message.content
        except Exception as e:
            msg = f'protocol_gpt: OpenAI API error: {e}'
            logger.error(msg)
            return f'<span style="color:red">{msg}</span>'




from importlib.metadata import version, PackageNotFoundError
import tomli
from pathlib import Path

try:
    __version__ = version("agi.green")
except PackageNotFoundError:
    # Fallback to reading pyproject.toml directly when in editable mode
    try:
        pyproject_path = Path(__file__).parent.parent / "pyproject.toml"
        with open(pyproject_path, "rb") as f:
            __version__ = tomli.load(f)["project"]["version"]
    except Exception:
        __version__ = "unknown"



from typing import Any, Callable
import hashlib
import asyncio
import weakref
import os

def hash_mutable(obj):
    return hashlib.sha256(str(obj).encode('utf-8')).hexdigest()

class DictNamespace(dict):
    '''An optionally asychronous reactive javascript style dict for attribute access to keys
    Attributes starting with _ are stored as actual attributes, not in the dict (unless written as dict keys) so they can be used for internal state
    Pre-existing attributes have priority, such as dict methods (e.g. copy, items, keys, ...)
    But these name may safely be accessed as dict keys, e.g. obj['items']=123, assert obj['items']==123, iter(obj.items())
    Collisions only occur when such attributes are explicitly set, e.g. obj.items=456 => ValueError

    Asynchronous only refers to the ability to bind change handlers to the object. If you don't use this feature, ignore async.

    Content is expected to be json serializable

    _default is the default value for missing attributes.
    Unlike defaultdict, the default value only applies to attribute access, not dict access

    _depth is the number of layers that are DictNamespace by default.
    _default is the default leaf value (or callable to generate it)
    note that unlike defaultdict, you don't need to wrap the default in a lambda

    You can make variable depth simply by assigning a DictNamespace()

    You can attach change handlers to the object with _bind_change_handler(handler).
    '''
    def __init__(self, _depth=0, _default: Any=None, **kwargs):
        super().__init__(**kwargs)

        if _depth > 0:
            self._default = lambda: DictNamespace(_depth=_depth-1, _default=_default)
        else:
            if isinstance(_default, Callable):
                self._default = _default
            else:
                self._default = lambda: _default

        self._dirty_hashes = {}

        if kwargs:
            self._deep_update(kwargs)

    def _deep_update(self, d):
        'update self with d recursively'
        for k, v in d.items():
            v0 = self.get(k, None)

            if isinstance(v, dict) and isinstance(v0, dict):
                if not isinstance(v0, DictNamespace):
                    v0 = DictNamespace(**v)
                    self[k]=v0

                v0._deep_update(v)
            else:
                if isinstance(v, dict) and not isinstance(v, DictNamespace):
                    v = DictNamespace(**v)

                self[k] = v

    def _changed(self, key:str):
        'return True if the object has been changed since the last time this was called with this key'
        h = hash_mutable(self)
        if h != self._dirty_hashes.get(key):
            self._dirty_hashes[key] = h
            return True
        return False

    def _ensure_finalization(self):
        'ensure that the object is _dead when it is garbage collected'
        if not hasattr(self, '_dead'):
            weakref.finalize(self, self._kill)
            self._dead = False

    def _kill(self):
        'kill the object'
        self._dead = True

    async def _bind_change_handler(self, handler:Callable, delay:float=0.1):
        'asyncronously call handler whenever this object changes, within the delay'
        key = id(handler)
        self._ensure_finalization()

        async def change_loop():
            while not self._dead:
                if self._changed(key):
                    await handler()
                await asyncio.sleep(delay)

        asyncio.create_task(change_loop())

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(name)

        try:
            return self[name]
        except KeyError:
            self[name] = self._default()
            return self[name]

    def _expandvars(self):
        'replace $VAR with os.environ["VAR"] for all strings, return the modified object'

        def expandvars(x):
            if isinstance(x, str):
                return os.path.expandvars(x)
            elif isinstance(x, dict):
                return DictNamespace(**{k: expandvars(v) for k, v in x.items()})
            elif isinstance(x, list):
                return [expandvars(v) for v in x]
            else:
                return x

        return expandvars(self)

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            try:
                super().__getattribute__(name)
                raise AttributeError(f"Cannot set predefined attribute `{name}`. Use obj['{name}'] instead, or use `_{name}` for internal state")
            except AttributeError:
                if isinstance(value, dict):
                    value = DictNamespace(**value)
                self[name] = value

    def __delattr__(self, name: str) -> None:
        if name.startswith('_'):
            super().__delattr__(name)
        else:
            del self[name]








from agi_green.dispatcher import Protocol, protocol_handler
import logging
import os

logger = logging.getLogger(__name__)
log_level = os.getenv('LOG_LEVEL', 'WARNING').upper()
logging.basicConfig(level=log_level)


class FormProtocol(Protocol):
    protocol_id = "form"

    @protocol_handler
    async def on_http_vueform_process(self, form_id: str, **kwargs) -> str:
        logger.info(f"vueform_process {form_id}: {kwargs}")
        await self.handle_mesg(form_id, **kwargs)
        return "OK"



import logging
from typing import Dict, Any
import aiohttp
from agi_green.dispatcher import Protocol, protocol_handler
import os
from aiohttp import web

logger = logging.getLogger(__name__)

class AzureProtocol(Protocol):
    """
    Azure protocol handler for managing Azure-specific user identification
    """
    protocol_id: str = 'azure'

    def __init__(self, parent: Protocol):
        super().__init__(parent)
        self.azure_user = None
        self.azure_id = None
        self.azure_tenant = None
        self.app_id = os.getenv('AZURE_APPLICATION_ID')
        self.tenant_id = os.getenv('AZURE_TENANT_ID')
        logger.info("AzureProtocol initialized")

    async def get_user_photo(self, access_token: str) -> str:
        """
        Fetch user's photo from Microsoft Graph API
        Returns photo URL or default avatar if unavailable
        """
        try:
            async with aiohttp.ClientSession() as session:
                headers = {'Authorization': f'Bearer {access_token}'}
                url = f'https://graph.microsoft.com/v1.0/me/photo/$value'
                async with session.get(url, headers=headers) as response:
                    if response.status == 200:
                        # Here you'd need to actually save the photo somewhere and return its URL
                        # For now, we'll just return the default
                        return '/avatars/azure.png'
                    else:
                        logger.error(f"Failed to fetch azure user photo: {response.status}")
        except Exception as e:
            logger.error(f"Failed to fetch user photo: {e}")

        return '/avatars/azure.png'

    @protocol_handler
    async def on_ws_connect(self, headers: Dict[str, str]):
        """
        Handle WebSocket connection and extract Azure user information from headers
        """
        self.azure_user = headers.get("X-MS-CLIENT-PRINCIPAL-NAME", None)
        self.azure_id = headers.get("X-MS-CLIENT-PRINCIPAL-ID", None)
        self.azure_tenant = headers.get("X-MS-CLIENT-PRINCIPAL-IDP", None)
        access_token = headers.get("X-MS-TOKEN-AAD-ACCESS-TOKEN", None)

        if self.azure_user and self.azure_id:
            logger.info(f"Azure User Connected - Name: {self.azure_user}, ID: {self.azure_id}, Tenant: {self.azure_tenant}")

            if access_token:
                icon = await self.get_user_photo(access_token)
            else:
                logger.warning("No access token found for Azure user")
                icon = '/avatars/azure.svg'

            await self.send('ws', 'set_user_data',
                            uid=self.azure_id,
                            name=self.azure_user,
                            icon=icon)

    @protocol_handler
    async def on_http_request(self, request_method: str, request_url: str, **kwargs):
        """Extract Azure headers from HTTP request"""
        headers = kwargs.get('headers', {})
        logger.info(f"HTTP Request headers: {headers}")

        # For local development, check if we're running outside Azure
        if not any(key.startswith('X-MS-CLIENT-') for key in headers):
            logger.debug("Running locally - checking for Azure auth endpoints")
            if '/.auth/login/aad' in request_url:
                # Redirect to Azure login
                auth_url = f"https://login.microsoftonline.com/{self.tenant_id}/oauth2/v2.0/authorize"
                params = {
                    'client_id': self.app_id,
                    'response_type': 'code',
                    'redirect_uri': 'http://localhost:8002/.auth/login/done',
                    'response_mode': 'query',
                    'scope': 'openid profile email'
                }
                redirect_url = f"{auth_url}?{'&'.join(f'{k}={v}' for k,v in params.items())}"
                return web.HTTPFound(redirect_url)
            elif '/.auth/login/done' in request_url:
                # Handle OAuth callback
                # This is simplified - in production you'd validate the token
                self.azure_user = "Local Dev User"
                self.azure_id = "local-dev-id"
                self.azure_tenant = self.tenant_id
                return web.HTTPFound('/')

        # Normal Azure App Service header processing
        self.azure_user = headers.get("X-MS-CLIENT-PRINCIPAL-NAME")
        self.azure_id = headers.get("X-MS-CLIENT-PRINCIPAL-ID")
        self.azure_tenant = headers.get("X-MS-CLIENT-PRINCIPAL-IDP")

        if any([self.azure_user, self.azure_id, self.azure_tenant]):
            logger.info(f"Azure headers found - User: {self.azure_user}, ID: {self.azure_id}, Tenant: {self.azure_tenant}")
        else:
            logger.debug("No Azure headers found in HTTP request")




from .dict_namespace import DictNamespace, hash_mutable
import yaml
import os
import shutil
import asyncio
import aiofiles
import yaml
import weakref

async def load_yaml(file_path) -> dict:
    async with aiofiles.open(file_path, 'r') as f:
        content = await f.read()
    try:
        r = yaml.safe_load(content)
    except:
        r = None
    return r

async def save_yaml(file_path, data:dict):
    async with aiofiles.open(file_path, 'w') as f:
        yaml_str = yaml.safe_dump(data)
        await f.write(yaml_str)

def multiline_str_representer(dumper: yaml.Dumper, data: str) -> yaml.ScalarNode:
    'because we like pretty multiline strings in our yaml config files'
    if isinstance(data, str) and '\n' in data:
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)

yaml.add_representer(str, multiline_str_representer)

class ConfigNamespace(DictNamespace):
    'Configuration namespace asynchonously mirrored to yaml'
    def __init__(self, filepath:str, default_filepath:str=None, depth=0, delay:float=0.1, writeable=False, **kwargs):
        '''filepath: path to yaml file or env var if prefixed with $
        default_filepath: path to copy to filepath if it doesn't exist
        delay: time to wait between checking for changes to the yaml file
        '''
        super().__init__(depth, **kwargs)
        self._filepath = os.path.expandvars(filepath)
        self._delay = delay
        self._writeable = writeable

        if not os.path.exists(self._filepath):
            # Create directory if it doesn't exist
            os.makedirs(os.path.dirname(self._filepath), exist_ok=True)

            # Copy from default if actual config doesn't exist
            if default_filepath:
                shutil.copy(default_filepath, self._filepath)
            else:
                # Create empty file if no default
                with open(self._filepath, 'w') as f:
                    f.write('{}')

        self._yaml_timestamp = 0 # last time the yaml file was read or written
        self._changed('yaml') # clear the change flag
        asyncio.create_task(self._sync_yaml_loop())

    def _yaml_changed(self):
        'return True if the yaml file has changed since last read'
        r = self._yaml_timestamp < os.path.getmtime(self._filepath)
        self._yaml_timestamp = os.path.getmtime(self._filepath)
        return r

    async def _sync_yaml(self):
        'sync yaml file to data and data to yaml file if changed. You may need to await this manually if you want to force a sync.'
        data_changed = self._writeable and self._changed('yaml')
        file_changed = self._yaml_changed()

        if file_changed:
            data = await load_yaml(self._filepath)
            if data is not None:
                if self == {}:
                    data_changed = False
                    self.clear()

                self._deep_update(data) # _deep_update ensures that DictNamespace objects are applied to depth

                self._changed('yaml') # clear the change

        if data_changed:
            await save_yaml(self._filepath, self)

    async def _sync_yaml_loop(self):
        'sync data to yaml file forever'
        self._ensure_finalization() # ensure task ends when we are garbage collected

        while not self._dead:
            await self._sync_yaml()
            await asyncio.sleep(self._delay)







'''cmd protocol

This protocol handles custom commands. Output is a string.

To invoke a command from a chat message, use the following syntax:

`[cmd:command_name(arg1=value1, arg2=value2, ...)]`

The command text is replaced with the result of the command.

To invoke a command programmatically, use the following syntax:

`result = await self.send('cmd', 'command_name', arg1=value1, arg2=value2, ...)`

And you can use the ws protocol to send the result to the chat:

`await self.send('ws', 'append_chat', author='info', message=result)`
'''

import os
from os.path import join, dirname, splitext, isabs
import re
from typing import Callable, Awaitable, Dict, Any, List, Set, Union, Tuple
from logging import getLogger, Logger
import logging
from os.path import exists
import ast

from agi_green.dispatcher import Protocol, format_call, protocol_handler

here = dirname(__file__)
logger = logging.getLogger(__name__)
log_level = os.getenv('LOG_LEVEL', 'WARNING').upper()
logging.basicConfig(level=log_level)

# [cmd:foo(x=5, y='333')]
# warning, this regex can't handle lists of tuples due to )] termination
# may be suplemented with extention to handle that
re_command = re.compile(r'''\[cmd:(\w+(?:[^\n](?!(?:\)\])))*.\))\]''')

def ast_node_to_value(node):
    ''
    if isinstance(node, ast.Constant):
        # Handle atomic literals like numbers, strings, etc.
        return node.value
    elif isinstance(node, ast.List):
        # Handle list literals
        return [ast_node_to_value(element) for element in node.elts]
    elif isinstance(node, ast.Tuple):
        # Handle tuple literals
        return tuple(ast_node_to_value(element) for element in node.elts)
    elif isinstance(node, ast.Dict):
        # Handle dict literals
        return {ast_node_to_value(key): ast_node_to_value(value) for key, value in zip(node.keys, node.values)}
    elif isinstance(node, ast.Set):
        # Handle set literals
        return {ast_node_to_value(element) for element in node.elts}
    # Add more cases here for other compound types if needed
    else:
        raise TypeError("Unsupported AST node type")

class CommandProtocol(Protocol):
    '''
    Command protocol

    Handle custom commands
    '''
    protocol_id: str = 'cmd'

    def __init__(self, parent:Protocol):
        super().__init__(parent)

    async def run(self):
        self.add_task(super().run())

    @protocol_handler(priority=1, update=True)
    async def on_ws_chat_input(self, content:str):
        'handle command syntax on chat input'

        # [cmd:gameio_start(game='y93', players=['user1', 'user2'])]

        for match in re_command.finditer(content):
            call_str = match.group(1)

            # TODO: Handle unbalanced '[(' due to regex limitations: tack on more '.*\)\]' matches
            # meanwhile lists of tuples will be broken
            # workaround is lists of lists.  This is a limitation of the regex, not the parser
            try:
                result = await self.send('cmd', call_str)
            except Exception as e:
                result = f'error: {e}'

            # replace the matched command with the result
            if result:
                content = content.replace(match.group(0), result)

        return {'content': content}

    async def do_send(self, cmd:str, **kwargs):
        '''cmd can be a function call expression like "foo(x=5, y='333')" or a simple function name like "foo"'''
        # Parse cmd as a function call expression using ast
        result = ''

        try:
            node = ast.parse(cmd, mode='eval').body

            if isinstance(node, ast.Name):
                return await super().do_send(cmd, **kwargs)

            elif isinstance(node, ast.Call):
                func_name = node.func.id
                kwargs |= {kw.arg: ast_node_to_value(kw.value) for kw in node.keywords}
                result = await self.send('cmd', func_name, **kwargs)
            else:
                result = f'error: Invalid command syntax: {cmd}'

        except (SyntaxError, ValueError) as e:
            # This might occur if the matched string isn't a valid Python function call
            result = f'error: {e} `{cmd}`'

        result = result or ''

        if result.startswith('error'):
            logger.error(result)
        else:
            logger.info('%s => "%s"', cmd, result)

        return result

